/*
 * Copyright (C) 2007-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file score.c Scoring functions to determine the audio file streaming order. */
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "list.h"

static struct osl_table *score_table;

static int ptr_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	void *d1 = *(void **)obj1->data;
	void *d2 = *(void **)obj2->data;
	return NUM_COMPARE(d1, d2);
}

/**
 * Compare the score of two audio files
 *
 * \param obj1 Pointer to the first score object.
 * \param obj2 Pointer to the second score object.
 *
 * This function first compares the score values as usual integers. If they compare as
 * equal, the address of \a obj1 and \a obj2 are compared. So this compare function
 * returns zero if and only if \a obj1 and \a obj2 point to the same memory area.
 *
 * \sa osl_compare_function.
 */
static int score_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	int d1 = *(int*)obj1->data;
	int d2 = *(int*)obj2->data;
	int ret = NUM_COMPARE(d2, d1);

	if (ret)
		return ret;
	return NUM_COMPARE(obj2->data, obj1->data);
}

/**
 * The score table consists of two columns: The \a aft_row column contains
 * pointers to the rows of the audio file table,  and the score column contains
 * the current score of the audio file associated with that row.
 */
enum score_table_columns {
	/** The row of the audio file. */
	SCORECOL_AFT_ROW,
	/** The score */
	SCORECOL_SCORE,
	/** This table has two columns */
	NUM_SCORE_COLUMNS
};

static struct osl_column_description score_cols[] = {
	[SCORECOL_AFT_ROW] = {
		.storage_type = OSL_NO_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE | OSL_DONT_FREE,
		.name = "aft_row",
		.compare_function = ptr_compare,
		.data_size = sizeof(void *)
	},
	[SCORECOL_SCORE] = {
		.storage_type = OSL_NO_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.name = "score",
		.compare_function = score_compare,
		.data_size = sizeof(long)
	}
};

static struct osl_table_description score_table_desc = {
	.name = "score",
	.num_columns = NUM_SCORE_COLUMNS,
	.flags = 0,
	.column_descriptions = score_cols
};

/**
 * Compute the number of files in score table.
 *
 * \param num Result is returned here.
 *
 * \return Positive on success, negative on errors.
 *
 * \sa osl_get_num_rows().
 */
int get_num_admissible_files(unsigned *num)
{
	return osl_get_num_rows(score_table, num);
}

/**
 * Get the score of the audio file associated with given row of the score table.
 *
 * \param score_row Pointer to the row in the score table.
 * \param score Result is returned here on success.
 *
 * On errors (negative return value) the content of \a score is undefined.
 *
 * \return The return value of the underlying call to osl_get_object().
 */
static int get_score_of_row(void *score_row, long *score)
{
	struct osl_object obj;
	int ret = osl_get_object(score_table, score_row, SCORECOL_SCORE, &obj);

	if (ret >= 0)
		*score = *(long *)obj.data;
	return ret;
}

/**
 * Add an entry to the table of admissible files.
 *
 * \param aft_row The audio file to be added.
 * \param score The score for this file.
 *
 * \return The return value of the underlying call to osl_add_row().
 */
int score_add(const struct osl_row *aft_row, long score)
{
	int ret;
	struct osl_object score_objs[NUM_SCORE_COLUMNS];
	size_t size;

	assert(aft_row);
	size = score_table_desc.column_descriptions[SCORECOL_AFT_ROW].data_size;
	score_objs[SCORECOL_AFT_ROW].data = (struct osl_row *)aft_row;
	score_objs[SCORECOL_AFT_ROW].size = size;

	size = score_table_desc.column_descriptions[SCORECOL_SCORE].data_size;
	score_objs[SCORECOL_SCORE].data = para_malloc(size);
	score_objs[SCORECOL_SCORE].size = size;
	*(int *)(score_objs[SCORECOL_SCORE].data) = score;

//	PARA_DEBUG_LOG("adding %p\n", *(void **) (score_objs[SCORECOL_AFT_ROW].data));
	ret = osl_add_row(score_table, score_objs);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		free(score_objs[SCORECOL_SCORE].data);
	}
	return ret;
}

static int get_nth_score(unsigned n, long *score)
{
	struct osl_row *row;
	int ret = osl_get_nth_row(score_table, SCORECOL_SCORE, n, &row);

	if (ret < 0)
		return ret;
	return get_score_of_row(row, score);
}

/**
 * Replace a row of the score table.
 *
 * \param aft_row Determines the audio file to change.
 * \param percent The position to re-insert the audio file.
 *
 * The percent parameter must be between \p 0 and 100 and. A value of zero
 * means to re-insert the audio file into the score table with a score lower
 * than any other admissible file.
 *
 * \return Positive on success, negative on errors. Possible errors: Errors
 * returned by osl_get_row(), get_num_admissible_files(), osl_get_nth_row(),
 * osl_get_object(), osl_update_object().
 */
int score_update(const struct osl_row *aft_row, long percent)
{
	struct osl_row *row;
	long new_score;
	unsigned n, new_pos;
	struct osl_object obj = {.data = (struct osl_row *)aft_row,
		.size = sizeof(aft_row)};
	int ret = osl_get_row(score_table, SCORECOL_AFT_ROW, &obj, &row);

	if (ret == -E_RB_KEY_NOT_FOUND) /* not an error */
		return 1;
	if (ret < 0)
		return ret;
	ret = get_num_admissible_files(&n);
	if (ret < 0)
		return ret;
	new_pos = 1 + (n - 1) * percent / 100;
	ret = get_nth_score(new_pos, &new_score);
	if (ret < 0)
		return ret;
	new_score--;
	obj.size = sizeof(long);
	obj.data = para_malloc(obj.size);
	*(long *)obj.data = new_score;
	PARA_DEBUG_LOG("new score: %ld, rank %u/%u\n", new_score, new_pos, n);
	return osl_update_object(score_table, row, SCORECOL_SCORE, &obj);
}

/**
 * Given an admissible file, get the corresponding row in the aft and the score.
 *
 * \param score_row Determines the admissible file.
 * \param score Result pointer.
 * \param aft_row Result pointer.
 *
 * \return Negative on errors, positive on success. Possible errors: Errors
 * returned by osl_get_object().
 */
int get_score_and_aft_row(struct osl_row *score_row, long *score,
		struct osl_row **aft_row)
{
	struct osl_object obj;
	int ret = get_score_of_row(score_row, score);

	if (ret < 0)
		return ret;
	ret = osl_get_object(score_table, score_row, SCORECOL_AFT_ROW, &obj);
	if (ret < 0)
		return ret;
	*aft_row = obj.data;
	return 1;
}

static int get_score_row_from_aft_row(const struct osl_row *aft_row,
		struct osl_row **score_row)
{
	struct osl_object obj = {.data = (struct osl_row *)aft_row,
		.size = sizeof(aft_row)};
	return osl_get_row(score_table, SCORECOL_AFT_ROW, &obj, score_row);

}

/**
 * Loop over all files in the score table.
 *
 * \param data A pointer to arbitrary data.
 * \param func Function to be called for each admissible file.
 *
 * \return The return value of the underlying call to osl_rbtree_loop().
 *
 * This is used for the ls command. The \a data parameter is passed as the
 * second argument to \a func.
 *
 * \sa admissible_file_loop_reverse().
 */
int admissible_file_loop(void *data, osl_rbtree_loop_func *func)
{
	return osl_rbtree_loop(score_table, SCORECOL_SCORE, data, func);
}

/**
 * Loop over all files in the score table in reverse order.
 *
 * \param data As in admissible_file_loop().
 * \param func As in admissible_file_loop().
 *
 * \return Same return value as admissible_file_loop().
 *
 * \sa admissible_file_loop(), osl_rbtree_loop_reverse().
 */
int admissible_file_loop_reverse(void *data, osl_rbtree_loop_func *func)
{
	return osl_rbtree_loop_reverse(score_table, SCORECOL_SCORE, data, func);
}

/**
 * Get the admissible audio file with highest score.
 *
 * \param aft_row Points to the row in the aft of the "best" audio file.
 * \param score Highest score value in the score table.
 *
 * \return Positive on success, negative on errors. Possible errors: Errors
 * returned by osl_rbtree_last_row(), osl_get_object().
 */
int score_get_best(struct osl_row **aft_row, long *score)
{
	struct osl_row *row;
	struct osl_object obj;
	int ret = osl_rbtree_last_row(score_table, SCORECOL_SCORE, &row);

	if (ret < 0)
		return ret;
	ret = osl_get_object(score_table, row, SCORECOL_AFT_ROW, &obj);
	if (ret < 0)
		return ret;
	*aft_row = obj.data;
	return get_score_of_row(row, score);
}

/**
 * Remove an entry from the rbtree of admissible files.
 *
 * \param aft_row The file which is no longer admissible.
 *
 * \return Positive on success, negative on errors. Possible errors:
 * Errors returned by osl_get_row() and osl_del_row().
 *
 * \sa score_add(), score_shutdown().
 */
int score_delete(const struct osl_row *aft_row)
{
	struct osl_row *score_row;
	int ret = get_score_row_from_aft_row(aft_row, &score_row);

	if (ret < 0)
		return ret;
	return osl_del_row(score_table, score_row);
}

/**
 * Find out whether an audio file is contained in the score table.
 *
 * \param aft_row The row of the audio file table.
 * \param rank Result pointer
 *
 * \return Positive, if \a aft_row belongs to the audio file table,
 * zero if not, negative on errors. If \a aft_row was found, and \a rank
 * is not \p NULL, the rank of \a aft_row is returned in \a rank.
 */
int row_belongs_to_score_table(const struct osl_row *aft_row, unsigned *rank)
{
	struct osl_row *score_row;
	int ret = get_score_row_from_aft_row(aft_row, &score_row);

	if (ret == -E_RB_KEY_NOT_FOUND)
		return 0;
	if (ret < 0)
		return ret;
	if (!rank)
		return 1;
	ret = osl_get_rank(score_table, score_row, SCORECOL_SCORE, rank);
	if (ret < 0)
		return ret;
	return 1;
}

/* Close the score table. */
static void score_close(void)
{
	osl_close_table(score_table, OSL_FREE_VOLATILE);
	score_table = NULL;
}

/**
 * Open the score table.
 *
 * \param dir Unused.
 *
 * \return The return value of the underlying call to osl_open_table().
 */
static int score_open(__a_unused const char *dir)
{
	score_table_desc.dir = NULL; /* this table has only volatile columns */
	return osl_open_table(&score_table_desc, &score_table);
}

/**
 * Remove all entries from the score table, but keep the table open.
 *
 * \return Standard.
 */
int clear_score_table(void)
{
	score_close();
	return score_open(NULL);
}

static int score_event_handler(__a_unused enum afs_events event,
		__a_unused struct para_buffer *pb, __a_unused void *data)
{
	return 1;
}

/**
 * Initialize the scoring subsystem.
 *
 * \param t The members of \a t are filled in by the function.
 */
void score_init(struct afs_table *t)
{
	t->name = score_table_desc.name;
	t->open = score_open;
	t->close = score_close;
	t->event_handler = score_event_handler;
}
